'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var util = require('util');
var _ = require('ramda');
var _$1 = require('ramda-adjunct');
require('lodash.merge');
var lodash = require('lodash');
var mapValues = require('lodash.mapvalues');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var mapValues__default = /*#__PURE__*/_interopDefaultLegacy(mapValues);

_.flip(_.concat);
const toList = _.pipe(_$1.list, _$1.flattenDepth(1));
_.map(toList);

_.curry((char, str) => _.pipe(_$1.trimCharsStart(char), _$1.trimCharsEnd(char))(str));

const prop = _.useWith(_.lensPath, [_.split('.')]);

const getProp = _.curryN(2, (key = '', value = {}) => {
    if (!_$1.isString(key) || !_$1.isPlainObject(value)) return value;

    return _.pipe(_.useWith(_.view, [prop]), _.defaultTo(value))(key, value);
});

const setProp = _.useWith(_.set, [prop]);

const $ = obj => ({
    pipe: operators => _.reduce(_.applyTo, obj, operators)
});

/* eslint-disable no-console */

const log = (..._) => {
    console.log(..._);
    return _[0];
};

const err = (..._) => {
    console.error(..._);
    return _[0];
};

const deep = (...args) => {
    const inspectOpts = {
        showHidden: false,
        depth: null,
        colors: true,
        maxArrayLength: null
    };
    const toPrint = _.intersperse(' ', args);
    const inspectOrNot = _ => _$1.isPlainObject(_) ? util.inspect(_, inspectOpts) : _;
    const formatted = _.map(inspectOrNot, toPrint);

    _.forEach(formatted, _ => process.stdout.write(_));
    console.log();

    return args[0];
};

const note = (...messages) => _ => {
    log(...messages);
    return _;
};
const label = (...messages) => _ => {
    log(...messages, _);
    return _;
};

log.err = err;
log.deep = deep;
log.note = note;
log.label = label;

// eslint-disable-next-line complexity
var log$1 = ((title, path, logger = log) => value => {
    if (!title) {
        logger(value);
        return Promise.resolve(value);
    }
    if (typeof title === 'function') {
        title(value);
        return Promise.resolve(value);
    }
    if (title && !path) {
        if (lodash.has(value, title)) {
            logger(getProp(title, value));
            return Promise.resolve(value);
        }
        logger(title, value);
        return Promise.resolve(value);
    }
    if (title && typeof path === 'function') {
        if (lodash.has(value, title)) {
            path(getProp(title, value));
            return Promise.resolve(value);
        }
        path(title, value);
        return Promise.resolve(value);
    }
    if (title && path) {
        if (lodash.has(value, path)) {
            logger(title, getProp(path, value));
            return Promise.resolve(value);
        }
        logger(title, value);
        return Promise.resolve(value);
    }
    return Promise.resolve(value);
});

// eslint-disable-next-line complexity
const merger = (obj, value) => {
    if (!value) return obj;
    if (!obj) return value;
    if (Array.isArray(obj) && Array.isArray(value)) return [...obj, ...value];
    if (Array.isArray(obj) && !Array.isArray(value)) return [...obj, value];
    if (_$1.isPlainObject(obj) && !_$1.isPlainObject(value)) return obj;
    if (_$1.isPlainObject(obj) && Array.isArray(value)) return obj;
    if (_$1.isPlainObject(obj) && _$1.isPlainObject(value)) return _.mergeWith(merger, obj, value);

    return obj;
};

var _extends = Object.assign || function (target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i];

    for (var key in source) {
      if (Object.prototype.hasOwnProperty.call(source, key)) {
        target[key] = source[key];
      }
    }
  }

  return target;
};

var slicedToArray = function () {
  function sliceIterator(arr, i) {
    var _arr = [];
    var _n = true;
    var _d = false;
    var _e = undefined;

    try {
      for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
        _arr.push(_s.value);

        if (i && _arr.length === i) break;
      }
    } catch (err) {
      _d = true;
      _e = err;
    } finally {
      try {
        if (!_n && _i["return"]) _i["return"]();
      } finally {
        if (_d) throw _e;
      }
    }

    return _arr;
  }

  return function (arr, i) {
    if (Array.isArray(arr)) {
      return arr;
    } else if (Symbol.iterator in Object(arr)) {
      return sliceIterator(arr, i);
    } else {
      throw new TypeError("Invalid attempt to destructure non-iterable instance");
    }
  };
}();

// eslint-disable-next-line complexity
var extend = ((path, obj) => snowball => {
    const value = _$1.isPlainObject(snowball) ? _extends({}, snowball) : snowball;
    // eslint-disable-next-line no-nested-ternary
    const paths = obj && Array.isArray(path) ? path : !path ? null : [path];
    const finalPath = obj ? paths : null;
    const base = obj || path;
    let finalObject = null;
    try {
        finalObject = typeof base === 'function' ? base(value) : base;
    } catch (e) {
        e.value = value;
        throw e;
    }
    if (finalPath) {
        return Promise.all(finalPath.map(single => Promise.resolve(finalObject).then(o => setProp(single.toString(), o, {})).then(o => merger(value, o)))).then(o => o.reduce(merger, {})).catch(e => {
            e.value = value;
            throw e;
        });
    }
    return Promise.resolve(finalObject).then(o => merger(value, o)).catch(e => {
        e.value = value;
        throw e;
    });
});

var extendEach = ((...args) => value => {
    var _$cond = _.cond([[_.equals(2), _.always([undefined, args[0], args[1]])], [_.equals(3), _.always(args)]])(args.filter(_.identity).length),
        _$cond2 = slicedToArray(_$cond, 3);

    const key = _$cond2[0],
          target = _$cond2[1],
          map = _$cond2[2];


    return $(value).pipe([getProp(target), _.ifElse(_$1.isPlainObject, _.mapObjIndexed((v, k) => map(v, value, k)), _$1.mapIndexed((v, k) => map(v, value, k))), _$1.list, _$1.flattenDepth(1), _.cond([[_.pipe(_.head, _$1.isPlainObject), _.reduce(merger, {})], [_.pipe(_.head, _$1.isArray), _.concat([])], [_.pipe(_.head, _.identity), _.identity]]), v => extend(key, v)(value)]);
});

var under = (path => value => Promise.resolve(setProp(path, value, {})));

var ifOperator = ((path, transducer) => value => {
    if (Array.isArray(path)) {
        if (Array.every(path, p => lodash.has(value, p))) return transducer(value);
        return Promise.resolve(value);
    }
    if (lodash.has(value, path)) return transducer(value);
    return Promise.resolve(value);
});

// eslint-disable-next-line complexity
var doOperator = ((path, sideEffect) => value => {
    const finalPath = sideEffect ? path : null;
    const finalSideEffect = sideEffect || path;
    const finalValue = finalPath ? getProp(finalPath, value) : value;
    let result = finalSideEffect;
    if (typeof finalSideEffect === 'function') result = finalSideEffect(finalValue);
    return Promise.resolve(result).then(() => value);
});

var ifNo = ((path, transducer) => value => {
    if (Array.isArray(path)) {
        if (Array.every(path, p => !lodash.has(value, p))) return transducer(value);
        return Promise.resolve(value);
    }
    if (!lodash.has(value, path)) return transducer(value);
    return Promise.resolve(value);
});

var switchOperator = ((path, branches) => value => {
    const test = getProp(path, value);
    if (lodash.has(branches, test)) return branches[test](value);
    // eslint-disable-next-line dot-notation
    if (lodash.has(branches, '_default')) return branches['_default'](value);
    return Promise.resolve(value);
});

var doEach = ((iterable, stream) => value => Promise.all(getProp(iterable, value).map(iterator => stream(value, iterator))).then(results => results.reduce((v, result) => doOperator(v)(result), value)));

var operators = {
    extend,
    extendEach,
    do: doOperator,
    doEach,
    if: ifOperator,
    ifNo,
    log: log$1,
    switch: switchOperator,
    under
};

var isolator = ((...args) => {
    const fn = path => value => {
        if (!Array.isArray(path)) return getProp(path, value);
        if (path.length === 0) return value;

        return path.reduce((result, currentPath) => {
            if (lodash.has(value, currentPath)) {
                return merger(result, setProp(currentPath, getProp(currentPath, value), {}));
            }
            return result;
        }, {});
    };
    if (args.length === 1) return fn(args[0]);
    return fn(args[0])(args[1]);
});

const createApi = ops => transducers => value => {
    if (value === undefined) {
        const pipeline = val => ({
            then: callback => transducers.reduce((prev, next) => prev.then(next), Promise.resolve(val)).then(callback)
        });
        return Object.assign(pipeline, mapValues__default['default'](ops, operator => (...args) => {
            transducers.push(operator(...args));
            return pipeline;
        }));
    }

    const api = mapValues__default['default'](ops, operator => (...args) => {
        transducers.push(operator(...args));
        return api;
    });
    api.then = callback => transducers.reduce((prev, next) => prev.then(next), Promise.resolve(value)).then(callback);
    return api;
};

var configure = (ops => value => createApi(ops)([])(value));

const transmuter = configure(operators);
Object.assign(transmuter, operators);
const configureTransmuter = configure;
const transmute = transmuter;
const isolate = isolator;

exports.configureTransmuter = configureTransmuter;
exports.default = transmuter;
exports.isolate = isolate;
exports.transmute = transmute;
